#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Generate the geometry of a rotor blade from parameters describing the cross-
section profiles and the stacking law of the blade.

[FR:
Generation de la geometrie d'une aube de turbomachine a partir de parametres
decrivant les profils des sections de l'aube et la loi d'empilement.
]

Copyright (c) 2019, Elsa Piollet (elsa.piollet@polymtl.ca)
                    Alain Batailly (alain.batailly@polymtl.ca)
                    Polytechnique Montreal, Canada

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
For details on license and warranty, see http://www.gnu.org/licenses
or LICENSE.txt.
"""

import os
import ConfigParser
import subprocess

from src.parameter_smoothing import parameter_smoothing
from src.profile_coordinates import profile_coordinates

def main():
    config = ConfigParser.ConfigParser()
    config.read("parameters.ini")

    smoothing = config.getboolean('main', 'smooth_parameters')
    coordinates = config.getboolean('main', 'compute_coordinates')
    cad = config.getboolean('main', 'compute_cad')

    current_path = os.path.dirname(os.path.realpath(__file__))

    if smoothing:
        # parameter smoothing
        # [FR: regularisation des parametres]
        print("Parameter smoothing in progress...")

        original_file = os.path.join(current_path,
                                     config.get('parameter_smoothing',
                                                'profile_folder'),
                                     config.get('parameter_smoothing',
                                                'original_filename'))
        smoothed_file = os.path.join(current_path,
                                     config.get('parameter_smoothing',
                                                'profile_folder'),
                                     config.get('parameter_smoothing',
                                                'smoothed_filename'))
        fig = config.getboolean('parameter_smoothing', 'display_fig')

        parameter_smoothing(original_file, smoothed_file, fig=fig)
        print("Parameter smoothing done.")

    if coordinates:
        # computation of the profile coordinates
        # [FR: calcul des coordonnees des profils]
        print("Profile coordinate computation in progress...")

        blade_name = config.get('profile_coordinates', 'blade_name')
        profile_file = os.path.join(current_path,
                                    config.get('profile_coordinates',
                                               'profile_folder'),
                                    config.get('profile_coordinates',
                                               'profile_filename'))

        sweep_lean = config.getboolean('profile_coordinates', 'sweep_lean')
        if sweep_lean:
            sweep_file = os.path.join(current_path,
                                      config.get('profile_coordinates',
                                                 'sweep_lean_folder'),
                                      blade_name + '_sweep.json')
            lean_file = os.path.join(current_path,
                                     config.get('profile_coordinates',
                                                'sweep_lean_folder'),
                                     blade_name + '_lean.json')
        else:
            sweep_file = ''
            lean_file = ''
        output_coor_directory = os.path.join(
            current_path, config.get('profile_coordinates',
                                    'output_folder'), blade_name)
        display_fig = config.getboolean('profile_coordinates', 'display_fig')

        # launch the main program
        profile_coordinates(profile_file, output_coor_directory,
                            sweep_file, lean_file, display_fig=display_fig,
                            save_profiles=True)

        print("Profile coordinate computation done.")

    if cad:
        # computation of the CAD and mesh in Salome
        # [FR: calcul de la CAO et du maillage dans Salome]
        print("CAD in progress...")
        salome_directory = config.get('blade_cad', 'salome_directory')
        subprocess.call("{} -t {}".format(
            salome_directory, os.path.join(
                current_path, 'src', 'blade_cad.py')), shell=True)
        print("CAD done.")


if __name__ == "__main__":
    main()