#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Compute the CAD and mesh of a compressor blade from cross-section
coordinates.
For more details, see README.md and docs/documentation.pdf

[FR:
Calcul de la CAO et du maillage d'une aube de compresseur à partir
des coordonees des sections transversales.
Pour plus de details, voir README.md et docs/documentation.pdf
(en anglais)
]

Copyright (c) 2019, Elsa Piollet (elsa.piollet@polymtl.ca)
                    Alain Batailly (alain.batailly@polymtl.ca)
                    Polytechnique Montreal, Canada

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
For details on license and warranty, see http://www.gnu.org/licenses
or LICENSE.txt.
"""

# built-in
import os
import inspect
import ConfigParser

# Salome libraries
import salome
import salome_notebook
from salome.geom import geomBuilder
import SMESH
from salome.smesh import smeshBuilder
from salome.StdMeshers import StdMeshersBuilder


def blade_cad():
    #
    # retrieve parameters from ini file
    # [FR: recuperation des parametres a partir du fichier ini]
    #

    # obtain the location of this file even when loaded directly from Salome
    # [FR: obtention de l'emplacement de ce fichier meme s'il est charge
    # depuis Salome]
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    script_path = os.path.dirname(os.path.abspath(filename))
    config = ConfigParser.ConfigParser()
    config.read(os.path.join(script_path, '..',"parameters.ini"))
    # general options [FR: options generales]
    blade_name = config.get('blade_cad', 'blade_name')
    # path to the profile coordinate folder
    # [FR: chemin vers le dossier des coordonnees de profils]
    profile_path = os.path.join(script_path, '..',
                                config.get('blade_cad','coordinate_folder'),
                                blade_name)
    # CAD options [FR: options de CAO]
    export_cad = config.getboolean('blade_cad', 'export_cad')
    cad_folder = os.path.join(script_path, '..', config.get('blade_cad', 'cad_folder'))
    # mesh options [FR: options de maillage]
    mesh = config.getboolean('blade_cad', 'compute_mesh')
    quadratic = config.getboolean('blade_cad', 'quadratic')
    number_segments = config.getint('blade_cad', 'number_segments')
    export_mesh = config.getboolean('blade_cad', 'export_mesh')
    mesh_folder = os.path.join(script_path, '..',
                               config.get('blade_cad', 'mesh_folder'),
                               blade_name)
    # profile orientation, (0, 1, 2) = (x, y, z)
    # [FR: orientation des profils, (0, 1, 2) = (x, y, z)]
    radial = 0
    tangent = 1
    axial = 2

    #
    # initialize Salome study
    # [FR: initialisation de l'etude Salome]
    #

    salome.salome_init()
    theStudy = salome.myStudy
    notebook = salome_notebook.NoteBook(theStudy)

    #
    # GEOM component [FR: composant GEOM]
    #

    geompy = geomBuilder.New(theStudy)

    O = geompy.MakeVertex(0, 0, 0)
    OX = geompy.MakeVectorDXDYDZ(1, 0, 0)
    OY = geompy.MakeVectorDXDYDZ(0, 1, 0)
    OZ = geompy.MakeVectorDXDYDZ(0, 0, 1)
    geompy.addToStudy(O, 'O')
    geompy.addToStudy(OX, 'OX')
    geompy.addToStudy(OY, 'OY')
    geompy.addToStudy(OZ, 'OZ')

    # list of available profile levels along the blade space
    # [FR: liste des niveaux de profils disponibles dans la hauteur de l'aube]
    level_list = []
    with open(os.path.join(profile_path, "percent_span.txt")) as file:
        for line in file:
            level_list.append(int(line))

    # two curves: suction side (s) and pressure side (p), obtained from the
    # corresponding text files for each profile level
    # [FR: deux courbes: extrados et intrados, obtenues a partir des fichiers
    # correspondants pour chaque niveau de profil]
    Curve_s=[]
    Curve_p=[]

    for idx, level in enumerate(level_list):
        with open(os.path.join(profile_path,
                               "profile_{}s.txt".format(level))) as file:
            pt = []
            for line in file:
                x, y, z = [float(v) for v in line.split()]
                pt.append(geompy.MakeVertex(x, y, z))

        Curve_s.append(geompy.MakeInterpol(pt,False))

        with open(os.path.join(profile_path,
                               "profile_{}p.txt".format(level))) as file:
            pt = []
            for line in file:
                x, y, z = [float(v) for v in line.split()]
                pt.append(geompy.MakeVertex(x, y, z))

        Curve_p.append(geompy.MakeInterpol(pt,False))

    # points on the trailing edge (te)
    # [FR: points sur le bord d'attaque]
    pt_te = []
    with open(os.path.join(profile_path,"profile_te.txt")) as file:
        coord = []
        for line in file:
            x, y, z = [float(v) for v in line.split()]
            coord.append([x,y,z])
            pt_te.append(geompy.MakeVertex(x, y, z))

    # list of Salome elements: edges and curves
    Edge_11 = [0]*len(level_list)
    Edge_12 = [0]*len(level_list)
    Edge_21 = [0]*len(level_list)
    Edge_22 = [0]*len(level_list)
    Cs = [0]*len(level_list)
    Cp = [0]*len(level_list)

    for idx in range(len(level_list)):
        Cs[idx] = geompy.DivideEdge(Curve_s[idx], -1, 0.5, 0)
        [Edge_11[idx], Edge_12[idx]] = geompy.ExtractShapes(
            Cs[idx], geompy.ShapeType["EDGE"], False)

        Cp[idx] = geompy.DivideEdge(Curve_p[idx], -1, 0.5, 0)
        [Edge_21[idx], Edge_22[idx]] = geompy.ExtractShapes(
            Cp[idx], geompy.ShapeType["EDGE"], False)

    Filling_11 = geompy.MakeFilling(Edge_11, isApprox=1)
    Filling_12 = geompy.MakeFilling(Edge_12, isApprox=1)
    Filling_21 = geompy.MakeFilling(Edge_21, isApprox=1)
    Filling_22 = geompy.MakeFilling(Edge_22, isApprox=1)
    Filling_1 = geompy.MakeFilling(Cs)
    Filling_2 = geompy.MakeFilling(Cp)
    Glue_1 = geompy.MakeGlueEdges([Filling_11, Filling_12, Filling_21, Filling_22],
                                  0.01)
    SuppressHoles_2 = geompy.SuppressHoles(Glue_1, [])
    Shell_2 = geompy.MakeShell([SuppressHoles_2])

    Blade = geompy.MakeSolid([Shell_2])
    geompy.addToStudy(Blade, 'Blade')

    [Face_tip,Face_hub] = geompy.SubShapes(Blade, [31, 33])
    geompy.addToStudyInFather(Blade, Face_hub, 'Face_hub')
    geompy.addToStudyInFather(Blade, Face_tip, 'Face_tip')

    if export_cad:
        if not os.path.exists(cad_folder):
            os.makedirs(cad_folder)
        geompy.ExportSTEP(Blade,os.path.join(cad_folder, blade_name + '.step'))

    if salome.sg.hasDesktop():
        salome.sg.updateObjBrowser(True)

    if mesh:
        ###
        ### Prepare necessary subshapes
        ###
        [Face_1, Face_2, Face_3, Face_4] = geompy.SubShapes(Blade, [3, 20, 13, 27])
        geompy.addToStudyInFather(Blade, Face_1, 'Face_1')
        geompy.addToStudyInFather(Blade, Face_2, 'Face_2')
        geompy.addToStudyInFather(Blade, Face_3, 'Face_3')
        geompy.addToStudyInFather(Blade, Face_4, 'Face_4')

        [Edge_1, Edge_2, Edge_3, Edge_4] = geompy.ExtractShapes(
            Face_1, geompy.ShapeType["EDGE"], True)
        [Edge_5, Edge_6, Edge_7, Edge_8] = geompy.ExtractShapes(
            Face_2, geompy.ShapeType["EDGE"], True)
        [Edge_9, Edge_10, Edge_11, Edge_12] = geompy.ExtractShapes(
            Face_3, geompy.ShapeType["EDGE"], True)
        [Edge_13, Edge_14, Edge_15, Edge_16] = geompy.ExtractShapes(
            Face_4, geompy.ShapeType["EDGE"], True)
        geompy.addToStudyInFather(Face_1, Edge_1, 'Edge_1')
        geompy.addToStudyInFather(Face_1, Edge_2, 'Edge_2')
        geompy.addToStudyInFather(Face_1, Edge_3, 'Edge_3')
        geompy.addToStudyInFather(Face_1, Edge_4, 'Edge_4')
        geompy.addToStudyInFather(Face_2, Edge_5, 'Edge_5')
        geompy.addToStudyInFather(Face_2, Edge_6, 'Edge_6')
        geompy.addToStudyInFather(Face_2, Edge_7, 'Edge_7')
        geompy.addToStudyInFather(Face_2, Edge_8, 'Edge_8')
        geompy.addToStudyInFather(Face_3, Edge_9, 'Edge_9')
        geompy.addToStudyInFather(Face_3, Edge_10, 'Edge_10')
        geompy.addToStudyInFather(Face_3, Edge_11, 'Edge_11')
        geompy.addToStudyInFather(Face_3, Edge_12, 'Edge_12')
        geompy.addToStudyInFather(Face_4, Edge_13, 'Edge_13')
        geompy.addToStudyInFather(Face_4, Edge_14, 'Edge_14')
        geompy.addToStudyInFather(Face_4, Edge_15, 'Edge_15')
        geompy.addToStudyInFather(Face_4, Edge_16, 'Edge_16')

        if salome.sg.hasDesktop():
            salome.sg.updateObjBrowser(True)

        ###
        ### SMESH component
        ###

        smesh = smeshBuilder.New(theStudy)
        Mesh_1 = smesh.Mesh(Blade)

        # discretization of the edges by number of segments
        # edges in the height of the blade are discretized in number_segments
        # while edges on the top and bottom which form the tip and hub are
        # discretized in half the number_segments as they
        # are smaller

        # depending on Salome version, the edge number may change: check in the GUI
        # which edge is which if issues arise
        # during meshing

        # height edges
        # Face 1
        Regular_1D = Mesh_1.Segment(geom=Edge_2)
        Number_of_Segments_1 = Regular_1D.NumberOfSegments(number_segments)
        Regular_1D_1 = Mesh_1.Segment(geom=Edge_3)
        status = Mesh_1.AddHypothesis(Number_of_Segments_1,Edge_3)
        # Face 2
        Regular_1D_2 = Mesh_1.Segment(geom=Edge_6)
        status = Mesh_1.AddHypothesis(Number_of_Segments_1,Edge_6)
        Regular_1D_3 = Mesh_1.Segment(geom=Edge_7)
        status = Mesh_1.AddHypothesis(Number_of_Segments_1,Edge_7)

        # top and bottom edges
        # Face 1
        Regular_1D_4 = Mesh_1.Segment(geom=Edge_1)
        Number_of_Segments_2 = Regular_1D_4.NumberOfSegments(number_segments/2)
        Regular_1D_5 = Mesh_1.Segment(geom=Edge_4)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_4)
        #Face 2
        Regular_1D_7 = Mesh_1.Segment(geom=Edge_5)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_5)
        Regular_1D_6 = Mesh_1.Segment(geom=Edge_8)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_8)
        # Face 3
        Regular_1D_8 = Mesh_1.Segment(geom=Edge_9)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_9)
        Regular_1D_9 = Mesh_1.Segment(geom=Edge_12)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_12)
        # Face 4
        Regular_1D_10 = Mesh_1.Segment(geom=Edge_13)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_13)
        Regular_1D_11 = Mesh_1.Segment(geom=Edge_16)
        status = Mesh_1.AddHypothesis(Number_of_Segments_2,Edge_16)

        Quadrangle_2D = Mesh_1.Quadrangle(algo=smeshBuilder.QUADRANGLE,
                                          geom=Face_1)
        Quadrangle_Parameters_1 = Quadrangle_2D.QuadrangleParameters(
            StdMeshersBuilder.QUAD_STANDARD,-1,[],[])
        Quadrangle_2D_1 = Mesh_1.Quadrangle(algo=smeshBuilder.QUADRANGLE,
                                            geom=Face_2)
        status = Mesh_1.AddHypothesis(Quadrangle_Parameters_1,Face_2)
        Quadrangle_2D_2 = Mesh_1.Quadrangle(algo=smeshBuilder.QUADRANGLE,
                                            geom=Face_3)
        status = Mesh_1.AddHypothesis(Quadrangle_Parameters_1,Face_3)
        Quadrangle_2D_3 = Mesh_1.Quadrangle(algo=smeshBuilder.QUADRANGLE,
                                            geom=Face_4)
        status = Mesh_1.AddHypothesis(Quadrangle_Parameters_1,Face_4)

        Sub_mesh_1 = Regular_1D.GetSubMesh()
        Sub_mesh_2 = Regular_1D_1.GetSubMesh()
        Sub_mesh_3 = Regular_1D_2.GetSubMesh()
        Sub_mesh_4 = Regular_1D_3.GetSubMesh()
        Sub_mesh_5 = Regular_1D_4.GetSubMesh()
        Sub_mesh_6 = Regular_1D_5.GetSubMesh()
        Sub_mesh_7 = Regular_1D_6.GetSubMesh()
        Sub_mesh_8 = Regular_1D_7.GetSubMesh()
        Sub_mesh_9 = Regular_1D_8.GetSubMesh()
        Sub_mesh_10 = Regular_1D_9.GetSubMesh()
        Sub_mesh_11 = Regular_1D_10.GetSubMesh()
        Sub_mesh_12 = Regular_1D_11.GetSubMesh()
        Sub_mesh_13 = Quadrangle_2D.GetSubMesh()
        Sub_mesh_14 = Quadrangle_2D_1.GetSubMesh()
        Sub_mesh_15 = Quadrangle_2D_2.GetSubMesh()
        Sub_mesh_16 = Quadrangle_2D_3.GetSubMesh()

        Projection_2D = Mesh_1.Projection2D(geom=Face_tip)
        Source_Face_1 = Projection_2D.SourceFace(Face_hub,None,None,None,None,None)

        MEFISTO_2D = Mesh_1.Triangle(algo=smeshBuilder.MEFISTO,geom=Face_hub)
        Length_From_Edges_1 = MEFISTO_2D.LengthFromEdges()

        Hexa_3D = Mesh_1.Hexahedron(algo=smeshBuilder.Hexa)

        Group_hub = Mesh_1.GroupOnGeom(Face_hub,'Group_1',SMESH.NODE)
        Group_tip = Mesh_1.GroupOnGeom(Face_tip,'Group_1',SMESH.NODE)

        isDone = Mesh_1.Compute()
        isDone = Mesh_1.SmoothParametricObject(
            Mesh_1, [], 20, 1.1, SMESH.SMESH_MeshEditor.LAPLACIAN_SMOOTH )

        if quadratic:
            Mesh_1.ConvertToQuadratic(0)

        if export_mesh:
            if not os.path.exists(mesh_folder):
                os.makedirs( mesh_folder)
            try:
                Mesh_1.ExportDAT(os.path.join(mesh_folder,'Mesh.dat'))
                pass
            except:
                print 'ExportDAT() failed. Invalid file name?'

            try:
                Mesh_1.ExportDAT(os.path.join(mesh_folder,'Group_hub.dat'),
                                 Group_hub)
                pass
            except:
                print 'ExportPartToDAT() failed. Invalid file name?'

            try:
                Mesh_1.ExportDAT(os.path.join(mesh_folder,'Group_tip.dat'),
                                 Group_tip )
                pass
            except:
                print 'ExportPartToDAT() failed. Invalid file name?'

        ## Set names of Mesh objects
        smesh.SetName(Mesh_1.GetMesh(), 'Mesh')
        smesh.SetName(Regular_1D.GetAlgorithm(), 'Regular_1D')
        smesh.SetName(Hexa_3D.GetAlgorithm(), 'Hexa_3D')
        smesh.SetName(Quadrangle_2D.GetAlgorithm(), 'Quadrangle_2D')

        smesh.SetName(MEFISTO_2D.GetAlgorithm(), 'MEFISTO_2D')
        smesh.SetName(Length_From_Edges_1, 'Length From Edges_1')
        smesh.SetName(Quadrangle_Parameters_1, 'Quadrangle Parameters_1')
        smesh.SetName(Projection_2D.GetAlgorithm(), 'Projection_2D')
        smesh.SetName(Source_Face_1, 'Source Face_1')
        smesh.SetName(Number_of_Segments_1, 'Number of Segments_1')
        smesh.SetName(Number_of_Segments_2, 'Number of Segments_2')

        smesh.SetName(Sub_mesh_1, 'Sub-mesh_1')
        smesh.SetName(Sub_mesh_2, 'Sub-mesh_2')
        smesh.SetName(Sub_mesh_3, 'Sub-mesh_3')
        smesh.SetName(Sub_mesh_4, 'Sub-mesh_4')
        smesh.SetName(Sub_mesh_5, 'Sub-mesh_5')
        smesh.SetName(Sub_mesh_6, 'Sub-mesh_6')
        smesh.SetName(Sub_mesh_7, 'Sub-mesh_7')
        smesh.SetName(Sub_mesh_8, 'Sub-mesh_8')
        smesh.SetName(Sub_mesh_9, 'Sub-mesh_9')
        smesh.SetName(Sub_mesh_10, 'Sub-mesh_10')
        smesh.SetName(Sub_mesh_11, 'Sub-mesh_11')
        smesh.SetName(Sub_mesh_12, 'Sub-mesh_12')
        smesh.SetName(Sub_mesh_13, 'Sub-mesh_13')
        smesh.SetName(Sub_mesh_14, 'Sub-mesh_14')
        smesh.SetName(Sub_mesh_15, 'Sub-mesh_15')
        smesh.SetName(Sub_mesh_16, 'Sub-mesh_16')

        smesh.SetName(Group_tip, 'Group_tip')
        smesh.SetName(Group_hub, 'Group_hub')


        if salome.sg.hasDesktop():
            salome.sg.updateObjBrowser(True)


if __name__ == "__main__":
    blade_cad()