#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Smooth multiple-circular-arc parameters to produce smoother
blade geometries.
For more details, see README.md and docs/documentation.pdf

[FR:
régularisation des parametres d'arcs circulaires multiples (MCA) pour
creer des geometries d'aubes plus regulieres.
Pour plus de details, voir README.md et docs/documentation.pdf
(en anglais)
]

Copyright (c) 2019, Elsa Piollet (elsa.piollet@polymtl.ca)
                    Alain Batailly (alain.batailly@polymtl.ca)
                    Polytechnique Montréal, Canada

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
For details on license and warranty, see http://www.gnu.org/licenses
or LICENSE.txt.
"""

# built-in [FR: modules Python]
import os
import ConfigParser
import inspect

# libraries [FR: modules externes]
import numpy as np
import matplotlib.pyplot as plt



def parameter_smoothing(original_file, smoothed_file, fig=False):
    """Smooth multiple-circular-arc parameters provided in a csv file with a
    cubic polynomical and write the smoothed parameters in a new csv file.

    [FR: regularisation des parametres d'arcs circulaires multiples fournis
    dans un fichier csv avec un polynome cubique et ecriture des parametres
    regularises dans un nouveau fichier csv]

    Parameters
        original_file (str): path to the original csv file
        smoothed_file (str): path to the csv file in which the smoothed
                parameters are written (all folders in the path should already
                exist)
        fig (bool): option to plot a figure comparing the original and smoothed
                parameters
    """

    (percent_span, ric, roc, kic, ktc, koc,
     ti, tm, to, zmc, ztc, zoc) = np.loadtxt(original_file, delimiter=',',
                                             skiprows=1, unpack=True)

    kic = np.radians(kic)
    ktc = np.radians(ktc)
    koc = np.radians(koc)

    alpha = np.arctan((roc - ric) / zoc)
    rtc = ric + ztc * np.tan(alpha)
    rmc = ric + zmc * np.tan(alpha)

    # obtain polynomial coefficients for a polynomial fit of the parameters
    # [FR: obtention des coefficients polynomiaux pour l'approximation des
    # parametres]
    pztc = np.poly1d(np.polyfit(rtc, ztc, 3))
    pzmc = np.poly1d(np.polyfit(rmc, zmc, 3))
    pzoc = np.poly1d(np.polyfit(roc, zoc, 3))
    pti = np.poly1d(np.polyfit(ric, ti, 3))
    ptm = np.poly1d(np.polyfit(rmc, tm, 3))
    pto = np.poly1d(np.polyfit(roc, to, 3))
    pkic = np.poly1d(np.polyfit(ric, kic, 3))
    pktc = np.poly1d(np.polyfit(rtc, ktc, 3))
    pkoc = np.poly1d(np.polyfit(roc, koc, 3))

    # plot the original parameters with respect to the radial position of the
    # points
    # [FR: representation graphique des parametres originaux par rapport a la
    # position radiale des points]
    if fig:
        fig0, ax0 = plt.subplots()
        ax0.plot(rtc, ztc, 'x-')
        ax0.plot(rmc, zmc, '*-')
        ax0.plot(roc, zoc, 'o-')

    # apply polynomial smoothing
    # [FR: application de la regularisation polynomiale]
    ztc = pztc(rtc)
    zmc = pzmc(rmc)
    zoc = pzoc(roc)
    kic = pkic(ric)
    ktc = pktc(rtc)
    koc = pkoc(roc)
    ti = pti(ric)
    tm = ptm(rmc)
    to = pto(roc)

    # compute dependent parameters
    # [FR: calcul des parametres dependants]
    roc = ric + zoc * np.tan(alpha)
    rtc = ric + ztc * np.tan(alpha)
    rmc = ric + zmc * np.tan(alpha)

    # plot the smoothed parameters with respect to the radial position of the
    # points
    # [FR: representation graphique des parametres regularises par rapport a la
    # position radiale des points]
    if fig:
        ax0.plot(rtc, ztc, 'x-')
        ax0.plot(rmc, zmc, '*-')
        ax0.plot(roc, zoc, 'o-')
        plt.show()

    # save in csv file
    # [FR: sauvegarde dans un fichier csv]
    table = np.zeros((len(percent_span), 12))
    table[:, 0] = percent_span
    table[:, 1] = ric
    table[:, 2] = roc
    table[:, 3] = np.degrees(kic)
    table[:, 4] = np.degrees(ktc)
    table[:, 5] = np.degrees(koc)
    table[:, 6] = ti
    table[:, 7] = tm
    table[:, 8] = to
    table[:, 9] = zmc
    table[:, 10] = ztc
    table[:, 11] = zoc
    np.savetxt(smoothed_file, table, delimiter=',', fmt='%2.3f',
               header=('percent span, ric, roc, kic, ktc, koc, ' +
                       'ti, tm, to, zmc, ztc, zoc'))


if __name__ == "__main__":
    #
    # retrieve parameters from ini file
    # [FR: recuperation des parametres a partir du fichier ini]
    #
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    script_path = os.path.dirname(os.path.abspath(filename))
    config = ConfigParser.ConfigParser()
    config.read(os.path.join(script_path, "..", "parameters.ini"))
    original_file = os.path.join(script_path, "..",
                                 config.get('parameter_smoothing',
                                            'profile_folder'),
                                 config.get('parameter_smoothing',
                                            'original_filename'))
    smoothed_file = os.path.join(script_path, "..",
                                 config.get('parameter_smoothing',
                                            'profile_folder'),
                                 config.get('parameter_smoothing',
                                            'smoothed_filename'))
    # figure allows comparing original and smoothed parameters
    # [FR: la representation graphique permet de comparer les parameters
    # originaux et regularises]
    fig = False

    parameter_smoothing(original_file, smoothed_file, fig=fig)