#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Compute the cross-section coordinates of a compressor blade composed
of multiple-circular-arc elements, including variable sweep and lean
along the blade span.
In the comments, equation references and original Fortran function names refer
to a NASA report from 1969 by Crouse: for more details, see README.md and
docs/documentation.pdf

[FR:
Calcul des coordonnees des sections transverses d'une aube de compresseur
composee d'elements de type arcs circulaires multiples, avec prise en compte
d'une fleche et d'une inclinaison variables selon la hauteur de l'aube.
Dans les commentaires, les references aux equations aux les noms originaux de
fonctions Fortran font reference a un rapport de la NASA de 1969 par Crouse:
pour plus de details, voir README.md et docs/documentation.pdf (en anglais)
]

Copyright (c) 2019, Elsa Piollet (elsa.piollet@polymtl.ca)
                    Alain Batailly (alain.batailly@polymtl.ca)
                    Polytechnique Montreal, Canada

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
For details on license and warranty, see http://www.gnu.org/licenses
or LICENSE.txt.
"""

# futures
from __future__ import division, print_function

# built-in
import os
import inspect
import ConfigParser

# libraries
import numpy as np
import scipy.integrate as integrate
import scipy.interpolate as si
from scipy.optimize import fsolve
import json
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

__author__ = "Elsa Piollet and Alain Batailly"
__contact__ = "elsa.piollet@polymtl.ca"
__copyright__ = "Copyright (c) 2019, Elsa Piollet and Alain Batailly," +\
                "Polytechnique Montréal"
__date__ = "2019/05/01"
__deprecated__ = False
__email__ = "elsa.piollet@polymtl.ca"
__license__ = "GPLv3"
__maintainer__ = "Elsa Piollet"
__status__ = "Development"
__version__ = "1.0"


def conical_coor(Rc, eps_c, kc, t, angle):
    """
    Calculate the conical coordinates of surface points at the ends of a
    thickness path of a profile.
    Original Fortran subroutine name: RAEP
    Refer to equations (16), (19), (21), and (22) for functional relations.

    Args:
        Rc (float): radial coordinate of centerline point (inlet or outlet)
        eps_c (float): angular coordinate of centerline point (inlet or outlet)
        kc (float): slope at centerline point (inlet or outlet)
        t (float): thickness at the end of the profile (inlet or outlet)
        angle (float): angular position of the point in the half circle closing
            the profile end

    Returns:
        R (float): radial coordinate of the point
        eps (float): angular coordinate of the point

    [FR:
    Calcul des coordonnees coniques des points de surface aux extremites
    de l'epaisseur d'un profil.
    Nom de la routine Fortran originale: RAEP
    Voir les equations (16), (19), (21) et (22) pour les relations
    fonctionnelles.

    Arguments:
        Rc (flottant): coordonnee radiale du point de la corde (bord d'attaque
            ou bord de fuite)
        eps_c (flottant): coordonnee angulaire du point de la corde (bord
            d'attaque ou bord de fuite)
        kc (flottant): pente du point de la corde (bord d'attaque ou bord
            de fuite)
        t (flottant): epaisseur a l'extremite du profil (bord d'attaque ou bord
            de fuite)
        angle (flottant): position angulaire du point d'interet sur le demi-
            cercle fermant l'extremite du profil

    Sortie:
        R (flottant): coordonnee R du point
        eps (flottant): coordonnee angulaire du point
    ]
    """

    k = kc + angle
    R = Rc + t/2 * np.cos(k)  # eq. (19)
    if abs(np.cos(k)) < 1e-8:
        eps = eps_c + np.sign(np.sin(k)) * t / 2 / Rc  # eq. (22)
    else:
        eps = eps_c + np.tan(k) * np.log(R/Rc)  # eq. (21)

    return R, eps


def pressure_suction(icase, Rtc, Rmc, Rip, Rop, Ris, Ros,
                     eps_tc, eps_ip, eps_op, eps_is, eps_os, kic, ktc, koc,
                     tm, zeta_ic, zeta_oc, Cic, Coc):
    """
    Compute the pressure and suction surfaces of a profile.
    Original Fortran function name: NEED

    Args:
        icase (int): case denoting whether transition point is at (1),
            ahead of (2), or behind (3) the maximum thickness point
        R(t,m)c (float): radial coordinate of the (transition point, maximum
            thickness point) of the centerline
        R(i,o)p (float): radial coordinate of the (inlet, outlet) of the pressure
            surface
        R(i,o)s (float): radial coordinate of the (inlet, outlet) of the suction
            surface
        eps_tc (float): angular coordinate of the transition point of the
            centerline
        eps_(i,o)p (float): angular coordinate of the (inlet,outlet) of the
            pressure surface
        eps_(i,o)s (float): angular coordinate of the (inlet,outlet) of the
            suction surface
        k(i,t,o)c (float): slope at the (inlet, transition point, outlet)
            of the centerline
        tm (float): thickness at the maximum thickness point
        zeta_(i,o)c (float): (inlet,outlet) segment constant of the centerline
        C(i,o)c (float): curvature constant at the (inlet,outlet) of the
            centerline

    Returns:
        R(t,m)p (float): radial coordinate of the (transition point, maximum
            thickness point) of the pressure surface
        R(t,m)s (float): radial coordinate of the (transition point, maximum
            thickness point) of the suction surface
        eps_mc (float): angular coordinate of the maximum thickness point of the
            centerline
        eps_(t,m)p (float): angular coordinate of the (transition point, maximum
            thickness point) of the pressure surface
        eps_(t,m)s (float): angular coordinate of the (transition point, maximum
            thickness point) of the suction surface
        k(i,t,o)p (float): slope at the (inlet, transition point, outlet)
            of the pressure surface
        k(i,t,o)s (float): slope at the (inlet, transition point, outlet)
            of the suction surface
        zeta_(i,o)p (float): (inlet,outlet) segment constant of the pressure
            surface
        zeta_(i,o)s (float): (inlet,outlet) segment constant of the suction
            surface
        C(i,o)p (float): curvature constant at the (inlet,outlet) of the
            pressure surface
        C(i,o)s (float): curvature constant at the (inlet,outlet) of the
            suction surface

    [FR:
    Calcul de l'intrados et de l'extrados d'un profil.
    Nom de la fonction Fortran originale: NEED

    Arguments:
        icase (entier): cas precisant si le point de transition est au meme
            endroit (1), avant (2) ou apres (3) le point d'epaisseur maximale
        R(t,m)c (flottant): coordonnee R au (point de transition, point
            d'epaisseur maximale) sur la corde
        R(i,o)p (flottant): coordonnee R au (bord d'attaque, bord de fuite) sur
            l'intrados
        R(i,o)s (flottant): coordonnee R au (bord d'attaque, bord de fuite) sur
            l'extrados
        eps_tc (flottant): coordonnee angulaire du point de transition sur la
            corde
        eps_(i,o)p (flottant): coordonnee angulaire du (bord d'attaque, bord de
            fuite) sur l'intrados
        eps_(i,o)s (flottant): coordonnee angulaire du (bord d'attaque, bord de
            fuite) sur l'extrados
        k(i,t,o)c (flottant): pente au (bord d'attaque, point de
            transition, bord de fuite) sur la corde
        tm (flottant): epaisseur au point d'epaisseur maximale
        zeta_(i,o)c (flottant): constante de segment au (bord d'attaque, bord de
            fuite) sur la corde
        C(i,o)c (flottant): constante de courbure au (bord d'attaque, bord de
            fuite) sur la corde

    Sortie:
        R(t,m)p (flottant): coordonnee radiale du (point de transition, point
            d'epaisseur maximale) sur l'intrados
        R(t,m)s (flottant): coordonnee radiale du (point de transition, point
            d'epaisseur maximale) sur l'extrados
        eps_mc (flottant): coordonnee angulaire du point d'epaisseur maximale
            sur la corde
        eps_(t,m)p (flottant): coordonnee angulaire du (point de transition,
            point d'epaisseur maximale) sur l'intrados
        eps_(t,m)s (flottant): coordonnee angulaire du (point de transition,
            point d'epaisseur maximale) sur l'extrados
        k(i,t,o)p (flottant): pente au (bord d'attaque, point de
            transition, bord de fuite) sur l'intrados
        k(i,t,o)s (flottant): pente au (bord d'attaque, point de
            transition, bord de fuite) sur l'extrados
        zeta_(i,o)p (flottant): constante de segment au (bord d'attaque, bord de
            fuite) sur l'intrados
        zeta_(i,o)s (flottant): constante de segment au (bord d'attaque, bord de
            fuite) sur l'extrados
        C(i,o)p (flottant): constante de courbure au (bord d'attaque, bord de
            fuite) sur l'intrados
        C(i,o)s (flottant): constante de courbure au (bord d'attaque, bord de
            fuite) sur l'extrados
    ]
    """

    if icase == 1:
        # transition at max. thickness
        # [FR: transition au point d'epaisseur maximale]
        km = ktc
        eps_mc = eps_tc
        Rms, eps_ms = conical_coor(Rmc, eps_mc, km, tm, np.pi / 2)
        Rmp, eps_mp = conical_coor(Rmc, eps_mc, km, tm, -np.pi / 2)
        ktp = ktc
        kts = ktc
        eps_tp = eps_mp
        eps_ts = eps_ms
        Rtp = Rmp
        Rts = Rms
        (kip, Cip, zeta_ip) = curve_equation(ktp, eps_tp, Rtp, eps_ip, Rip, kic)
        (kis, Cis, zeta_is) = curve_equation(kts, eps_ts, Rts, eps_is, Ris, kic)
        (kop, Cop, zeta_op) = curve_equation(ktp, eps_tp, Rtp, eps_op, Rop, koc)
        (kos, Cos, zeta_os) = curve_equation(kts, eps_ts, Rts, eps_os, Ros, koc)
    elif icase == 2:
        # transition ahead of max. thickness
        # [FR: transition avant le point d'epaisseur maximale]
        # eq. (23), max thickness on outlet segment
        # [FR: eq. (23), epaisseur maximale sur le segment du bord de fuite]
        km = np.arcsin(zeta_oc - Coc*Rmc)
        # eq. (11)
        eps_mc = eps_tc + epsilon_variation(km, ktc, zeta_oc, Rmc, Rtc)
        Rms, eps_ms = conical_coor(Rmc, eps_mc, km, tm, np.pi / 2)
        Rmp, eps_mp = conical_coor(Rmc, eps_mc, km, tm, -np.pi / 2)
        (kop, Cop, zeta_op) = curve_equation(km, eps_mp, Rmp, eps_op, Rop, koc)
        (kos, Cos, zeta_os) = curve_equation(km, eps_ms, Rms, eps_os, Ros, koc)
        (ktp, Rtp, eps_tp) = transition_point(km, Rmp, eps_mp, zeta_op,
                                              Cop, ktc, Rtc, eps_tc)
        (kts, Rts, eps_ts) = transition_point(km, Rms, eps_ms, zeta_os,
                                              Cos, ktc, Rtc, eps_tc)
        (kip, Cip, zeta_ip) = curve_equation(ktp, eps_tp, Rtp, eps_ip, Rip, kic)
        (kis, Cis, zeta_is) = curve_equation(kts, eps_ts, Rts, eps_is, Ris, kic)
    elif icase == 3:
        # transition behind max. thickness
        # [FR: transition apres le point d'epaisseur maximale]
        sinkm = zeta_ic - Cic * Rmc
        km = np.arcsin(sinkm)
        eps_mc = eps_tc + epsilon_variation(km, ktc, zeta_ic, Rmc, Rtc)
        Rms, eps_ms = conical_coor(Rmc, eps_mc, km, tm, np.pi / 2)
        Rmp, eps_mp = conical_coor(Rmc, eps_mc, km, tm, -np.pi / 2)
        (kip, Cip, zeta_ip) = curve_equation(km, eps_mp, Rmp, eps_ip, Rip, kic)
        (kis, Cis, zeta_is) = curve_equation(km, eps_ms, Rms, eps_is, Ris, kic)
        (ktp, Rtp, eps_tp) = transition_point(km, Rmp, eps_mp, zeta_ip, Cip,
                                              ktc, Rtc, eps_tc)
        (kts, Rts, eps_ts) = transition_point(km, Rms, eps_ms, zeta_is, Cis,
                                              ktc, Rtc, eps_tc)
        (kop, Cop, zeta_op) = curve_equation(ktp, eps_tp, Rtp, eps_op, Rop, koc)
        (kos, Cos, zeta_os) = curve_equation(kts, eps_ts, Rts, eps_os, Ros, koc)
    else:
        raise ValueError("unknown case")
    return (Rtp, Rmp, Rts, Rms, eps_mc, eps_tp, eps_mp, eps_ts, eps_ms,
            kip, ktp, kop, kis, kts, kos, zeta_ip, zeta_op, zeta_is, zeta_os,
            Cip, Cop, Cis, Cos)


def curve_equation(k1, eps_1, R1, eps_2, R2, ke):
    """
    Solve the equation of a constant dk/ds curve which passes through two known
    points and at a given slope at one of the points. Refer to equations (7),
    (11), and (15) for the functional relations.
    Original Fortran subroutine name: ITER

    Args:
        k1 (float): slope at point 1
        eps_1 (float): angular coordinate of point 1
        R1 (float): radial coordinate of point 1
        eps_2 (float): angular coordinate of point 2
        R2 (float): radial coordinate of point 2
        ke (float): an initial estimate of k2

    Returns:
        k2 (float): slope at point 2
        C(k2 (float)): curvature constant
        zeta(k2) (float): segment constant

    [FR:
    Resolution de l'equation d'une courbe de courbure constante (dk/ds
    constante) qui passe par deux point connus et dont la pente est connue en
    l'un des points.
    Voir les equations (7), (11) et (15) pour les relations fonctionnelles.
    Nom de la routine Fortran originale: ITER

    Arguments:
        k1 (flottant): pente au point 1
        eps_1 (flottant): coordonnee angulaire du point 1
        R1 (flottant): coordonnee raidale au point 1
        eps_2 (flottant): coordonnee angulaire du point 2
        R2 (flottant): coordonnee radiale du point 2
        ke (flottant): estimation initiale de k2

    Sortie:
        k2 (flottant): pente au point 2
        C (flottant): constante de courbure
        zeta (flottant): constante de segment
    ]
    """

    # eq (15)
    def C(k): return (np.sin(k1) - np.sin(k)) / (R2 - R1)

    # eq (7)
    def zeta(k): return C(k) * R1 + np.sin(k1)

    def F(k): return eps_2 - eps_1 - epsilon_variation(k, k1, zeta(k), R2, R1)

    k2 = fsolve(F, ke)

    return k2, C(k2), zeta(k2)


def transition_point(km, Rm, eps_m, zeta, C, ktc, Rtc, eps_tc):
    """
    Solve for the R-epsilon coordinates of the transition point on either the
    pressure surface or the suction surface.
    Refer to equations (11), (21) to (23) and (25) for the functional relations.
    Original Fortran subroutine name: ITER1

    Args:
        km (float): slope at maximum thickness point
        Rm (float): radial coordinate of the maximum thickness point
        eps_m (float): angular coordinate of the maximum thickness point
        zeta (float): segment constant
        C (float): curvature constant
        ktc (float): slope at the transition point of the centerline
        Rtc (float): radial coordinate of the transition point of the centerline
        eps_tc (float): angular coordinate of the transition point of the
            centerline

    Returns:
        kt (float): slope at the transition point
        Rt (float): radial coordinate of the transition point
        eps_t (float): angular coordinate of the transition point

    [FR:
    Calcul des coordonnees R-epsilon du point de transition sur l'intrados ou
    l'extrados.
    Voir les equations (11), (21) a (23) et (25) pour les relations
    fonctionnelles.
    Nom de la routine Fortran originale: ITER1

    Arguments:
        km (flottant): pente au point d'epaisseur maximale
        Rm (flottant): coordonnee radiale du point d'epaisseur maximale
        eps_m (flottant): coordonnee angulaire du point d'epaisseur maximale
        zeta (flottant): constante de segment
        C (flottant): constante de courbure
        ktc (flottant): pente au point de transition sur la corde
        Rtc (flottant): coordonnee radiale du point de transition sur la corde
        eps_tc (flottant): coordonnee angulaire du point de transition sur
            la corde

    Sortie:
        kt (flottant): pente au point de transition
        Rt (flottant): coordonnee radiale du point de transition
        eps_t (flottant): coordonnee angulaire du point de transition
    ]
    """
    # equation (23) + (7)
    kt = lambda R: fix_arcsin(np.sin(km) + C * (Rm - R))
    # equation (11)
    eps_t = lambda R: eps_m + epsilon_variation(kt(R), km, zeta, R, Rm)
    # equation (25)
    Rtc2 = lambda R: R / np.exp(np.tan(ktc) * (eps_tc - eps_t(R)))
    F = lambda R: Rtc - Rtc2(R)
    Rt = fsolve(F, Rtc)

    return kt(Rt), Rt, eps_t(Rt)


def epsilon_variation(k, k1, zeta, R, R1):
    """
    Compute the variation of angular epsilon coordinate between a point of
    interest and a reference point by evaluating the function f(k,k1,zeta,R,R1)
    as given in appendix B
    Original Fortran function name: SUBF

    Args:
        k (float): slope of curve at the point of interest
        k1 (float): slope of the curve at a reference point
        zeta (float): segment constant
        R (float): radial coordinate of the point of interest
        R1 (float): radial coordinate of a reference point

    Returns:
        f (float): value of the function f(k,k1,zeta,R,R1) as given in
            appendix B

    [FR:
    Calcul de la variation de la coordonnee angulaire epsilon entre un point
    d'interet et un point de reference en evaluant la fonction f(k,k1,zeta,R,R1)
    telle que donnee en annexe B.
    Nom de la fonction Fortran originale: SUBF

    Arguments:
        k (flottant): pente de la courbe au point d'interet
        k1 (flottant): point de la courbe au point de reference
        zeta (flottant): constante de segment
        R (flottant): coordonnee radiale du point d'interet
        R1 (flottant): coordonnee radiale du point de reference

    Sortie:
        f (flottant): valeur de la fonction f(k,k1,zeta,R,R1) telle que donnee
            en annexe B
    ]
    """

    if k == k1:
        f = np.tan(k)*np.log(R/R1)
    else:
        if zeta**2 > 1:
            # eq. (B9) + correction sqrt
            f = (k - k1 +
                 2*zeta/np.sqrt(zeta**2-1) *
                     (np.arctan((1-zeta*np.tan(k/2))/np.sqrt(zeta**2-1)) -
                      np.arctan((1-zeta*np.tan(k1/2))/np.sqrt(zeta**2-1))))
        elif zeta**2 < 1:
            # eq. (B10)
            f = (k - k1 +
                 zeta/np.sqrt(1-zeta**2) *
                     (np.log(abs((1 - zeta*np.tan(k/2) - np.sqrt(1 - zeta**2)) /
                                 (1 - zeta*np.tan(k/2) + np.sqrt(1 - zeta**2)))) -
                      np.log(abs((1 - zeta*np.tan(k1/2) - np.sqrt(1-zeta**2)) /
                                 (1 - zeta*np.tan(k1/2) + np.sqrt(1-zeta**2))))))
        elif zeta == 1:
            # eq. (B11)
            f = k - k1 + (np.tan(np.pi/4 + k1/2) - np.tan(np.pi/4 + k/2))
        else:  # zeta = -1
            # eq. (B11)
            f = k - k1 - (np.tan(np.pi/4 - k1/2) - np.tan(np.pi/4 - k/2))
    return f


def area_integrands(R, Rtp, Rts, eps_tp, eps_ts, ktp, kts,
                    zeta_ip, zeta_op, zeta_is, zeta_os, Cip, Cop, Cis, Cos):
    """
    Define the functions to integrate in function area_integration (below)
    Original Fortran subroutine name: part of INTGR

    Args:
        R (float): radial coordinate, integration variable
        Rtp (float): radial coordinate of the transition point of the
            pressure surface
        Rts (float): radial coordinate of the transition point of the
            suction surface
        eps_tp (float): angular coordinate of the transition point of the
            pressure surface
        eps_ts (float): angular coordinate of the transition point of the
            suction surface
        ktp (float): slope at the transition point of the pressure surface
        kts (float): slope at the transition point of the suction surface
        zeta_(i,o)p (float): segment constant at the (inlet, outlet) of the
            pressure surface
        zeta_(i,o)s (float): segment constant at the (inlet, outlet) of the
            suction surface
        C(i,o)p (float): curvature constant at the (inlet, outlet) of the
            pressure surface
        C(i,o)s (float): curvature constant at the (inlet, outlet) of the
            suction surface

    Returns:
        fint_dA (float): integrand for profile area computation (eq. (28))
        fint_RdA (float): integrand for integration of R over the profile area
            (eq. (29))
        fint_edA (float): integrand for integration of epsilon over the profile
            area (eq. (30))


    [FR:
    Definition des fonctions a integrer dans la fonction area_integration
    (ci-dessous).
    Nom de la routine Fortran originale: partie de INTGR

    Arguments:
        R (flottant): coordonnee radiale, variable d'integration
        Rtp (flottant): coordonnee radiale du point de transition de l'intrados
        Rts (flottant): coordonnee radiale du point de transition de l'extrados
        eps_tp (flottant): coordonnee angulaire du point de transition de
            l'intrados
        eps_ts (flottant): coordonnee angulaire du point de transition de
            l'extrados
        ktp (flottant): pente du point de transition de l'intrados
        kts (flottant): pente du point de transition de l'extrados
        zeta_(i,o)p (flottant): constante de segment au (bord d'attque, bord de
            fuite) de l'intrados
        zeta_(i,o)s (flottant): constante de segment au (bord d'attque, bord de
            fuite) de l'extrados
        C(i,o)p (flottant): constante de courbure au (bord d'attque, bord de
            fuite) de l'intrados
        C(i,o)s (flottant): constante de courbure au (bord d'attque, bord de
            fuite) de l'extrados

    Sortie:
        fint_dA (flottant): fonction a integrer pour le calcul de l'aire du
            profil (eq. (28))
        fint_RdA (flottant): fonction a integrer pour l'integration de R sur la
            surface du profil (eq. (29))
        fint_edA (flottant): fonction a integrer pour l'integration de epsilon
            sur la surface du profil (eq. (30))
    ]
    """

    if R > Rts:
        # eq. (23)
        kos = np.arcsin(zeta_os - R * Cos)
        # eq. (11)
        eps_s = eps_ts + epsilon_variation(kos, kts, zeta_os, R, Rts)
    else:
        # eq. (23)
        kis = np.arcsin(zeta_is - R * Cis)
        # eq. (11)
        eps_s = eps_ts + epsilon_variation(kis, kts, zeta_is, R, Rts)
    if R > Rtp:
        # eq. (23)
        kop = np.arcsin(zeta_op - R * Cop)
        # eq. (11)
        eps_p = eps_tp + epsilon_variation(kop, ktp, zeta_op, R, Rtp)
    else:
        # eq. (23)
        kip = np.arcsin(zeta_ip - R * Cip)
        # eq. (11)
        eps_p = eps_tp + epsilon_variation(kip, ktp, zeta_ip, R, Rtp)
    # eq. (28)
    fint_dA = R * (eps_s - eps_p)
    # eq. (29)
    fint_RdA = R ** 2 * (eps_s - eps_p)
    # eq. (30)
    fint_edA = R * (eps_s ** 2 - eps_p ** 2)/2
    return fint_dA, fint_RdA, fint_edA


def area_integration(Ric, Roc, Rtp, Rts, eps_tp, eps_ts, ktp, kts,
                     zeta_ip, zeta_op, zeta_is, zeta_os,
                     Cip, Cop, Cis, Cos):
    """
    Evaluate three definite integrals defined by equations (28), (29), and (30)
    to evaluate the R-epsilon coordinates of the center of gravity of the
    profile.
    Original Fortran subroutine name: part of INTGR

    Args:
        R(i,o)c (float): radial coordinate of the (inlet, outlet) of the
            centerline
        Rtp (float): radial coordinate of the otransition point of the
            pressure surface
        Rts (float): radial coordinate of the transition point of the
            suction surface
        eps_tp (float): angular coordinate of the transition point of the
            pressure surface
        eps_ts (float): angular coordinate of the transition point of the
            suction surface
        ktp (float): slope at the transition point of the pressure surface
        kts (float): slope at the transition point of the suction surface
        zeta_(i,o)p (float): segment constant at the (inlet, outlet) of the
            pressure surface
        zeta_(i,o)s (float): segment constant at the (inlet, outlet) of the
            suction surface
        C(i,o)p (float): curvature constant at the inlet of the pressure surface
        C(i,o)s (float): curvature constant at the outlet of the suction surface

    Returns:
        int_dA (float): profile area (eq. (28))
        int_RdA (float): integration of R over the profile area (eq. (29))
        int_edA (float): integration of epsilon over the profile area (eq. (30))

    [FR:
    Evaluation de trois integrales definies par les equations (28), (29), et
    (30) pour evaluer les coordonnees R-epsilon du centre de gravite du profil.
    Nom de la routine Fortran originale: partie de INTGR

    Arguments:
        R(i,o)c (flottant): coordonnee radiale du (bord d'attaque, bord de
            fuite) sur la corde
        Rtp (flottant): coordonnee radiale du point de transition de l'intrados
        Rts (flottant): coordonnee radiale du point de transition de l'extrados
        eps_tp (flottant): coordonnee angulaire du point de transition de
            l'intrados
        eps_ts (flottant): coordonnee angulaire du point de transition de
            l'extrados
        ktp (flottant): pente du point de transition de l'intrados
        kts (flottant): pente du point de transition de l'extrados
        zeta_(i,o)p (flottant): constante de segment au (bord d'attque, bord de
            fuite) de l'intrados
        zeta_(i,o)s (flottant): constante de segment au (bord d'attque, bord de
            fuite) de l'extrados
        C(i,o)p (flottant): constante de courbure au (bord d'attque, bord de
            fuite) de l'intrados
        C(i,o)s (flottant): constante de courbure au (bord d'attque, bord de
            fuite) de l'extrados

    Sortie:
        int_dA (flottant): aire du profil (eq. (28))
        int_RdA (flottant): integration de R sur la surface du profil
            (eq. (29))
        int_edA (flottant): integration de epsilon sur la surface du profil
            (eq. (30))
    ]

    """

    int_dA = integrate.quad(lambda R: area_integrands(
        R, Rtp, Rts, eps_tp, eps_ts, ktp, kts,
        zeta_ip, zeta_op, zeta_is, zeta_os, Cip, Cop, Cis, Cos)[0], Ric, Roc)[0]
    int_RdA = integrate.quad(lambda R: area_integrands(
        R, Rtp, Rts, eps_tp, eps_ts, ktp, kts,
        zeta_ip, zeta_op, zeta_is, zeta_os, Cip, Cop, Cis, Cos)[1], Ric, Roc)[0]
    int_edA = integrate.quad(lambda R: area_integrands(
        R, Rtp, Rts, eps_tp, eps_ts, ktp, kts,
        zeta_ip, zeta_op, zeta_is, zeta_os, Cip, Cop, Cis, Cos)[2], Ric, Roc)[0]

    return int_dA, int_RdA, int_edA


def interpolation_extrapolation(x, y, x0):
    """
    Compute y0 ordinate at x0 abscissa by interpolation (spline) or
    extrapolation (linear) of abscissa vector x and ordinate vector y.
    Original Fortran subroutine name: SINTP

    Args:
        x (array): abscissa vector
        y (array): corresponding ordinate vector
        x0 (float): abscissa of interest

    Returns:
        y0 (float): interpolated/extrapolated ordinate (i.e. y0 = y(x0))

    [FR:
    Calcul de l'ordonnee y0 a l'abscisse x0 par interpolation (spline) ou
    extrapolation (lineaire) des vecteurs d'abscisses et d'ordonnees x et y.
    Nom de la routine Fortran originale: SINTP

    Arguments:
        x (vecteur): vecteur d'abscisses
        y (vecteur): vecteur d'ordonnees correspondant
        x0 (flottant): abscisse d'interet

    Sortie:
        y0 (flottant): ordonnee interpolee/extrapolee (i.e. y0 = y(x0))
    ]
    """

    # sort x, and rearrange y to keep (x,y) correspondance
    # [FR: tri de x et reorganisation de y pour grader la correspondance (x,y)]
    ind = x.argsort()
    x = x[ind]
    y = y[ind]

    if x0 < x[0]:
        return y[0] + (x0 - x[0]) * (y[1] - y[0]) / (x[1] - x[0])
    elif x0 > x[-1]:
        return y[-1] + (x0 - x[-1]) * (y[-1] - y[-2]) / (x[-1] - x[-2])
    else:
        # spline interpolation
        # [FR: interpolation par une spline]
        (t, c, k) = si.splrep(x, y)
        y0 = si.splev(x0, (t, c, k))
        return y0


def fix_arcsin(a):
    """
    Compute the arcsin of a and correct for absolute values of a exceeding 1.
    Original Fortran subroutine name: FIX

    Args:
        a (float): argument of the arcsin

    Returns:
        b (float): arcsin value corrected for absolute values of a exceeding 1

    [FR:
    Calcul du arcsin de a, avec correction lorsque la valeur absolue de a est
    superieure a 1.
    Nom de la routine Fortran originale: FIX

    Arguments:
        a (flottant): argument du arcsin

    Sortie:
        b (flottant): valeur de arcsin corrigee pour les valeurs absolues de a
            superieures a 1
    ]
    """

    if abs(a)<1:
        b = np.arcsin(a)
    else:
        b = np.sign(a)*np.arcsin(1)
    return b


def xyz_coord(R, eps, alpha, theta_sp, Rref, zref, delta):
    """
    Compute cartesian coordinates of a point from its R-epsilon coordinates,
    R and z coordinates of a reference point, and information on the profile
    (cone angle, stacking point, lean)

    Args:
        R (float): radial coordinate of the point of interest
        eps (float): angular coordinate of the point of interest
        alpha (float): cone angle of the profile
        theta_sp (float): theta coordinate of the stacking point
        Rref (float): radial coordinate of the reference point
        zref (float): z-coordinate of the reference point
        delta (float): theta-coordinate variation due to lean

    Returns:
        x, y, z (float): cartesian coordinates of the point of interest

    [FR:
    Calcul des coordonnees cartesiennes d'un point a partir de ses coordonnees
    R-epsilon, des coordonnees R-z d'un point de reference, et d'informations
    sur le profil (angle du cone, point d'empilement, inclinaison).

    Arguments:
        R (flottant): coordonnee radiale du point d'interet
        eps (flottant): coordonnee angulaire du point d'interet
        alpha (flottant): angle de cone du profil
        theta_sp (flottant): coordonnee theta du point d'empilement
        Rref (flottant): coordonnee radiale du point de reference
        zref (flottant): coordonnee axiale du point de reference
        delta (flottant): variation de la coordonnee theta due a l'inclinaison

    Sortie:
        x, y, z (flottant): coordonnees cartesiennes du point d'interet
    ]
    """

    # eq. (36) (with theta_ic - theta_sph = delta - theta_sp)
    x = R * np.sin(alpha) * np.cos(eps / np.sin(alpha) + delta - theta_sp)
    # eq. (37) (with theta_ic - theta_sph = delta - theta_sp)
    y = R * np.sin(alpha) * np.sin(eps / np.sin(alpha) + delta - theta_sp)
    # eq. (38) (generalized to any reference point)
    z = zref - (Rref - R) * np.cos(alpha)
    return x, y, z


def center_gravity(yipx, zipx, yisx, zisx, ypx, ysx, zx, yopx,
                   zopx, yosx, zosx, nint=200):
    """
    Compute the center of gravity of a blade section (horizontal section at
    constant x coordinate)
    Original Fortran subroutine name: CGS

    Args:
        yipx (float): y-coordinate of the inlet of the pressure surface for the
            blade section
        zipx (float): z-coordinate of the inlet of the pressure surface for the
            blade section
        yisx (float): y-coordinate of the inlet of the suction surface for the
            blade section
        zisx (float): z-coordinate of the inlet of the suction surface for the
            blade section
        ypx (array): vector of y-coordinates of the pressure surface
            corresponding to z-coordinates in zx vector
        ysx (array): vector of y-coordinates of the suction surface
            corresponding to z-coordinates in zx vector
        zx (array): vector of z-coordinates
        yopx (float): y-coordinate of the outlet of the pressure surface for the
            blade section
        zopx (float): z-coordinate of the outlet of the pressure surface for the
            blade section
        yosx (float): y-coordinate of the outlet of the suction surface for the
            blade section
        zosx (float):  z-coordinate of the outlet of the suction surface for the
            blade section
        nint (int): number of integration points

    Returns:
        ycg (float): y-coordinate of the center of gravity of the blade section
        zcg (float): z-coordinate of the center of gravity of the blade section


    [FR:
    Calcul du centre de gravite d'une section d'aube (section horizontale prise
    a une coordonnee x constante).
    Nom de la routine Fortran originale: CGS.

    Arguments:
        yipx (flottant): coordonnee y de l'extremite amont de l'intrados pour la
            section d'aube
        zipx (flottant): coordonnee z de l'extremite amont de l'intrados pour la
            section d'aube
        yisx (flottant): coordonnee y de l'extremite amont de l'extrados pour la
            section d'aube
        zisx (flottant): coordonnee z de l'extremite amont de l'extrados pour la
            section d'aube
        ypx (vecteur): vecteur des coordonnees y de l'intrados correspondant aux
            coordonnees z du vecteur zx
        ysx (vecteur): vecteur des coordonnees y de l'extrados correspondant aux
            coordonnees z du vecteur zx
        zx (vecteur): vecteur de coordonnees z
        yopx (flottant): coordonnee y de l'extremite aval de l'intrados pour la
            section d'aube
        zopx (flottant): coordonnee z de l'extremite aval de l'intrados pour la
            section d'aube
        yosx (flottant): coordonnee y de l'extremite aval de l'extrados pour la
            section d'aube
        zosx (flottant):  coordonnee z de l'extremite aval de l'extrados pour la
            section d'aube
        nint (entier): nombre de points d'integration

    Sortie:
        ycg (flottant): coordonnee y du centre de gravite de la section d'aube
        zcg (flottant): coordonnee z du centre de gravite de la section d'aube
    ]
    """
    v = np.zeros(nint)
    v1 = np.zeros(nint)
    v2 = np.zeros(nint)
    yp = np.zeros(nint)
    ys = np.zeros(nint)

    dyi = yisx - yipx
    dzi = zisx - zipx
    dyo = yosx - yopx
    dzo = zosx - zopx

    ti2 = dyi**2 + dzi**2
    to2 = dyo**2 + dzo**2

    tix = np.sqrt(ti2)
    tox = np.sqrt(to2)

    sinbi = -dzi/tix
    sinbo = -dzo/tox
    cosbi = dyi/tix
    cosbo = dyo/tox

    if dzi >= 0:
        zmin = zipx
        zmin2 = zisx
    else:
        zmin = zisx
        zmin2 = zipx

    if dzo >= 0:
        zmax = zosx
        zmax2 = zopx
    else:
        zmax = zopx
        zmax2 = zosx

    z = np.linspace(zmin, zmax, nint)

    for i in range(nint):
        if z[i] >= zmin2:
            if z[i] > zmax2:
                if dzo >= 0:
                    yp[i] = yopx + (z[i] - zopx) * dyo/dzo
                    ys[i] = interpolation_extrapolation(zx, ysx, z[i])
                else:
                    ys[i] = yosx + (z[i] - zosx) * dyo/dzo
                    yp[i] = interpolation_extrapolation(zx, ypx, z[i])
            else:
                ys[i] = interpolation_extrapolation(zx, ysx, z[i])
                yp[i] = interpolation_extrapolation(zx, ypx, z[i])
        else:
            if dzi >= 0:
                ys[i] = yipx + (z[i] - zipx) * dyi/dzi
                yp[i] = interpolation_extrapolation(zx, ypx, z[i])
            else:
                yp[i] = yisx + (z[i] - zisx) * dyi/dzi
                ys[i] = interpolation_extrapolation(zx, ysx, z[i])
        v[i] = ys[i]-yp[i]
        v1[i] = (ys[i]**2-yp[i]**2)/2
        v2[i] = z[i]*(ys[i]-yp[i])

    Ai = ti2 * np.pi / 8
    Ao = to2 * np.pi / 8

    ybi = (yisx + yipx) / 2 - 2 / 3 / np.pi * tix * sinbi
    ybo = (yosx + yopx) / 2 + 2 / 3 / np.pi * tox * sinbo

    zbi = (zisx + zipx) / 2 - 2 / 3 / np.pi * tix * cosbi
    zbo = (zosx + zopx) / 2 + 2 / 3 / np.pi * tox * cosbo

    A = np.trapz(v, x=z) + Ai + Ao

    ycg = (np.trapz(v1, x=z) + ybi*Ai + ybo*Ao)/A  # eq. (45)

    zcg = (np.trapz(v2, x=z) + zbi*Ai + zbo*Ao)/A  # eq. (46)

    return ycg, zcg


def profile_coordinates(profile_csv, output_coor, sweep_json='', lean_json='',
                        display_fig=False, nx=10, nz=150,
                        save_profiles=False):
    """
    Compute the geometry of a compressor blade from multiple-cicular-arc
    parameters and sweep and lean variation along the blade span. The geometry
    is expressed as a point cloud with cartesian coordinates, stored in text
    files organized as follows:
    - two files for each percent span (hub - 0 to tip - 100):
       - pressure surface points
       - suction surface points
    - one file for the leading edges along the blade span
    - one file for the trailing edges along the blade span
    - one file listing the percent spans

    Args:
        profile_csv (str): path to the csv file containing multiple-circular-arc
            parameters
        output_coor (str): path to the folder in which the coordinate files are
            saved
        sweep_json (str): path to the json file containg sweep parameters
        lean_json (str): path to the json file containg lean parameters
        display_fig (bool): option to plot a 3D figure showing the resulting
            blade geometry
        nx (int): number of sampled horizontal blade sections in the x direction
            (for interpolation purposes)
        nz (int): number of sampling points in the z (axial) direction (applied
            to the profile with largest z-length)
        save_profiles (bool): option to write the computed coordinates to
            .txt files

    [FR:
    Calcul de la geometrie d'une aube de compresseur a partir de parametres
    d'arcs circulaires multiples et de la variation de la fleche et de
    l'inclinaison dans la hauteur de l'aube. La geometrie est exprimee comme
    un nuage de point avec des coordonnees cartesiennes, ecrites dans des
    fichiers textes organises comme suit:
    - deux fichiers pour chaque pourcentage de hauteur (du pied - 0 au
    sommet - 100):
       - points de l'intrados
       - points de l'extrados
    - un fichier pour les bords d'attaque le long de la hauteur de l'aube
    - un fichier pour les bords de fuite le long de la hauteur de l'aube
    - un fichier listant les pourcentages de hauteur

    Arguments:
        profile_csv (str): chemin vers le fichier csv contenant les parametres
            d'arcs circulaires multiples
        output_coor (str): chemin vers le dossier dans lequel les fichiers de
            coordonnees sont sauvegardes
        sweep_json (str): chemin vers le fichier json contenant les parametres
            de fleche
        lean_json (str): chemin vers le fichier json contenant les parametres
            d'inclinaison
        display_fig (bool): option de representer graphiquement la geometrie
            3D de l'aube
        nx (entier): nombre de sections horizontales dans la direction x
            (pour interpolation)
        nz (entier): nombre de points d'echantillonnage dans la direction z
            (axiale), nombre applique au profil le plus long selon z
        save_profiles (bool): option de sauvegarder les coordonnees calculees
            dans des fichiers .txt
    ]
    """

    # load multiple-circular-arc profile parameters
    # [FR: chargement des parametres de profils en arcs circulaires multiples]
    (percent_span, ric, roc, kic, ktc, koc, ti, tm, to,
     zmc, ztc, zoc) = np.loadtxt(profile_csv, delimiter=',',
                                 skiprows=1, unpack=True)

    # correction in angle sign for consistency between provided values in
    # reports and the actual compressor geometry of rotor 37
    # [FR: correction du signe de l'angle par coherence entre les valeurs
    # fournies dans les rapports et la geometrie reelle du rotor 37]
    kic = -np.radians(kic)
    ktc = -np.radians(ktc)
    koc = -np.radians(koc)

    # number of parameterized profiles in the radial direction
    # [FR: nombre de profils parametres dans la direction radiale]
    nprofiles = len(percent_span)
    
    # hub profile index: the program accepts parameters organized from hub to
    # tip or from tip to hub, 0 being always the hub
    # [FR: indice du pied d'aube: le programme accepte les parametres organises
    # du pied au sommet ou du sommet au pied, 0 etant toujours le pied]
    hub = np.where(percent_span == 0)

    # load 3D angle parameters
    # [FR: chargement des parametres 3D]
    if lean_json:
        with open(lean_json) as lean_file:
            lean = json.load(lean_file)
        # interpolation
        (t, c, k) = si.splrep(lean['percent_span'], lean['values'])
        lean['interp_val'] = si.splev(percent_span, (t, c, k))
        if lean['type'] == 'angle':
            lean['interp_val'] = np.radians(lean['interp_val'])
            if lean['angle_ref'] == 'hub':
                # original Crouse notation
                # [FR: notation originale de Crouse]
                eta = lean['interp_val']

    if sweep_json:
        with open(sweep_json) as sweep_file:
            sweep = json.load(sweep_file)
        (t, c, k) = si.splrep(sweep['percent_span'], sweep['values'])
        sweep['interp_val'] = si.splev(percent_span, (t, c, k))
        if sweep['type'] == 'angle':
            sweep['interp_val'] = np.radians(sweep['interp_val'])
            if sweep['angle_ref'] == 'hub':
                # original Crouse notation
                lam = sweep['interp_val']

    ## initialization
    # indices:
    # c: centerline, p: pressure side, s: suction side; i: inlet/leading edge,
    # o: outlet/trailing edge, t: transition point, m: max thickness point
    # [FR:
    # indices:
    # c: corde, p: intrados, s: extrados, i: amont/bord d'attaque,
    # o: aval/bord de fuite, t: point de transition, m: point d'epaisseur
    # maximale]

    # cartesian coordinates
    # [FR: coordonnees cartesiennes]
    (xic, xis, xip, xoc, xos, xop, yic, yis, yip, yoc, yos, yop,
     zic, zis, zipp, zos, zop) = (np.zeros(nprofiles) for i in range(17))
    (xtc, xts, xtp, xmc, xms, xmp, ytc, yts, ytp, ymc, yms, ymp,
     zts, ztp, zms, zmp) = (np.zeros(nprofiles) for i in range(16))
    yipx, yisx, yicx, yopx, yosx, yocx = (np.zeros(nx) for i in range(6))
    zipx, zisx, zicx, zopx, zosx, zocx = (np.zeros(nx) for i in range(6))
    xc, yc, zc, xp, yp, xs, ys = (np.zeros((nprofiles, nz)) for i in range(7))
    kappa_c, kappa_p, kappa_s, epsilon_c, epsilon_p, epsilon_s =\
        (np.zeros((nprofiles, nz)) for i in range(6))
    # radial coordinates R
    # [FR: coordonnees radiales R]
    Ric, Rip, Ris, Roc, Rop, Ros, Rtc, Rtp, Rts, Rmc, Rmp, Rms =\
        (np.zeros(nprofiles) for i in range(12))
    # angular coordinates epsilon
    # [FR: coordonnees angulaires epsilon]
    (eps_ic, eps_ip, eps_is, eps_oc, eps_op, eps_os, eps_tc,
     eps_tp, eps_ts, eps_mc, eps_mp, eps_ms) =\
        (np.zeros(nprofiles) for i in range(12))
    # segment constant (eq. (7))
    # [FR: constantes de segment (eq. (7))]
    zeta_ic, zeta_ip, zeta_is, zeta_oc, zeta_op, zeta_os =\
        (np.zeros(nprofiles) for i in range(6))
    # local blade angle
    # [FR: angle d'aube local]
    kip, kis, kop, kos, ktp, kts, kmc, kmp, kms =\
        (np.zeros(nprofiles) for i in range(9))
    # turning-rate constant
    # [FR: constante de courbure]
    Cic, Cip, Cis, Coc, Cop, Cos, Ctc, Ctp, Cts, Cmc, Cmp, Cms =\
        (np.zeros(nprofiles) for i in range(12))
    # case denoting whether transition point is at (1),
    # ahead of (2), or behind (3) the maximum thickness point
    # [FR: cas precisant si le point de transition est au meme
    # endroit (1), avant (2) ou apres (3) le point d'epaisseur maximale]
    icase = np.zeros(nprofiles)
    # stacking point
    # [FR: point d'emilement]
    xsp, ysp, zsp, Rsp, rsp, theta_sp = (np.zeros(nprofiles) for i in range(6))
    theta_ic = np.zeros(nprofiles)  #
    ycx, ypx, ysx = (np.zeros((nx, nz)) for i in range(3))
    yspx, zspx = (np.zeros(nx) for i in range(2))
    xspr, yspr, zspr = (np.zeros(nprofiles) for i in range(3))
    delta = np.zeros(nprofiles)

    ## begining of the computation
    ## [FR: debut du calcul]
    alpha = np.arctan((roc - ric) / zoc)

    ## centerline
    for i in range(nprofiles):
        # R coordinate for special points
        # [FR: coordonnee radiale pour des points speciaux]
        Ric[i] = ric[i] / np.sin(alpha[i])
        # equation (14)
        Roc[i] = Ric[i] + zoc[i] / np.cos(alpha[i])
        Rtc[i] = Ric[i] + ztc[i] / np.cos(alpha[i])
        Rmc[i] = Ric[i] + zmc[i] / np.cos(alpha[i])

        # obtain constant rate C from the blade angles at the beginning and end
        # of the inlet and outlet circular arcs
        # [FR: obtention de la constante de courbure C a partir des angles
        # de debut et de fin des segments d'attaque et de fuite]
        # equation (15)
        Cic[i] = (np.sin(kic[i]) - np.sin(ktc[i])) / (Rtc[i] - Ric[i])
        Coc[i] = (np.sin(koc[i]) - np.sin(ktc[i])) / (Rtc[i] - Roc[i])

        if ztc[i] == zmc[i]:
            kmc[i] = ktc[i]
            icase[i] = 1
        elif ztc[i] < zmc[i]:
            kmc[i] = koc[i]
            icase[i] = 2
        elif ztc[i] > zmc[i]:
            kmc[i] = kic[i]
            icase[i] = 3

        ## suction side and pressure side
        ## [FR: extrados et intrados]
        (Ris[i], eps_is[i]) = conical_coor(Ric[i], eps_ic[i], kic[i],
                                           ti[i], np.pi/2)
        (Rip[i], eps_ip[i]) = conical_coor(Ric[i], eps_ic[i], kic[i],
                                           ti[i], -np.pi/2)
        # corresponding constants zeta
        # [FR: constantes zeta correspondantes]
        zeta_ic[i] = Cic[i] * Rtc[i] + np.sin(ktc[i])  # eq. (7) and (B12)
        zeta_oc[i] = Coc[i] * Rtc[i] + np.sin(ktc[i])  # eq. (7)
        # angular coordinates
        # [FR: coordonnees angulaires]
        eps_ic[i] = 0
        # eq. (11)
        eps_tc[i] = epsilon_variation(ktc[i], kic[i], zeta_ic[i],
                                      Rtc[i], Ric[i])
        # eq. (11)
        eps_oc[i] = eps_tc[i] + epsilon_variation(koc[i], ktc[i], zeta_oc[i],
                                                  Roc[i], Rtc[i])
        (Ros[i], eps_os[i]) = conical_coor(Roc[i], eps_oc[i], koc[i],
                                           to[i], np.pi/2)
        (Rop[i], eps_op[i]) = conical_coor(Roc[i], eps_oc[i], koc[i],
                                           to[i], -np.pi/2)

        (Rtp[i], Rmp[i], Rts[i], Rms[i], eps_mc[i], eps_tp[i], eps_mp[i],
         eps_ts[i], eps_ms[i],
         kip[i], ktp[i], kop[i], kis[i], kts[i], kos[i], zeta_ip[i], zeta_op[i],
         zeta_is[i], zeta_os[i],
         Cip[i], Cop[i], Cis[i], Cos[i]) = pressure_suction(
            icase[i], Rtc[i], Rmc[i], Rip[i], Rop[i], Ris[i], Ros[i],
            eps_tc[i], eps_ip[i], eps_op[i], eps_is[i], eps_os[i],
            kic[i], ktc[i], koc[i], tm[i],
            zeta_ic[i], zeta_oc[i], Cic[i], Coc[i])

        ## step 1 - initial positioning of blade elements along stacking line
        # [FR: etape 1 - positionnement initial des profils le long de la ligne
        # d'empilement]
        int_dA, int_RdA, int_edA = area_integration(
            Ric[i], Roc[i], Rtp[i], Rts[i], eps_tp[i], eps_ts[i],
            ktp[i], kts[i], zeta_ip[i], zeta_op[i], zeta_is[i], zeta_os[i],
            Cip[i], Cop[i], Cis[i], Cos[i])

        # eq. (26) and (34)
        theta_sp[i] = int_edA / int_dA / np.sin(alpha[i])
        # eq. (26)
        Rsp[i] = int_RdA / int_dA
        # eq. (33)
        rsp[i] = Rsp[i] * np.sin(alpha[i])

    ## iterations for alignment
    # [FR: iterations pour l'alignement]
    tolerance = 0.001
    niter = 25
    n = 0
    success = False
    last_step = False

    while n < niter and not success:
        ## step 2 - calculation of stacking points relative to a
        # common reference
        ## [FR: etape 2 - calcul des points d'empilement sur base d'une
        # reference commune]
        theta_sph = theta_sp[hub]
        zsph = (Rsp[hub] - Ric[hub]) * np.cos(alpha[hub])
        rsph = rsp[hub]

        for i in range(nprofiles):
            if last_step and lean_json:
                if lean['type'] == 'angle':
                    if lean['angle_ref'] == 'hub':
                        # original Crouse formulation
                        # [FR: formulation originale de Crouse]
                        # eq. (35)
                        delta[i] = np.arcsin(
                            (rsph / rsp[i]) * np.tan(eta[i]) /
                            (1 + np.tan(eta[i]) ** 2) *
                            (np.sqrt((rsp[i] / rsph) ** 2 *
                                     (1 + np.tan(eta[i]) ** 2) -
                                     np.tan(eta[i]) ** 2) - 1))
                    elif lean['angle_ref'] == 'rotor_axis':
                        delta[i] = lean['interp_val'][i]
                elif lean['type'] == 'disp':
                    delta[i] = np.arcsin(lean['interp_val'][i]/rsp[i])
            else:
                delta[i] = 0
            if last_step and sweep_json:
                if sweep['type'] == 'angle':
                    if sweep['angle_ref'] == 'hub':
                        # original Crouse formulation
                        # [FR: formulation originale de Crouse]
                        # eq. (31)
                        zsp[i] = zsph + (rsp[i] - rsph) * np.tan(lam[i])
                    elif sweep['angle_ref'] == 'rotor_axis':
                        zsp[i] = zsph + rsp[i] * np.tan(sweep['interp_val'][i])
                elif sweep['type'] == 'disp':
                    zsp[i] = zsph + sweep['interp_val'][i]
            else:
                zsp[i] = zsph
            # if no lean or sweep, xsp is equal to the radial position
            # [FR: en l'absence d'inclinaison ou de fleche, xsp est egal a la
            # position radiale]
            xsp[i] = rsp[i] * np.cos(delta[i])
            theta_ic[i] = theta_sph + delta[i] - theta_sp[i]

            ## step 3 - computation of blade element coordinates
            # -> conversion of (R,epsilon) to (x,y,z)
            ## [FR: etape 3 - calcul des coordonnees des profils
            # -> conversion de (R,epsilon vers (x,y,z)) ]

            temp, temp, zic[i] = xyz_coord(Ric[i], 0, alpha[i],
                                           0, Rsp[i], zsp[i], 0)
            xip[i], yip[i], zipp[i] = xyz_coord(Rip[i], eps_ip[i], alpha[i],
                                                theta_sp[i], Ric[i], zic[i],
                                                delta[i])
            xis[i], yis[i], zis[i] = xyz_coord(Ris[i], eps_is[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xic[i], yic[i], zic[i] = xyz_coord(Ric[i], eps_ic[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xtp[i], ytp[i], ztp[i] = xyz_coord(Rtp[i], eps_tp[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xts[i], yts[i], zts[i] = xyz_coord(Rts[i], eps_ts[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xtc[i], ytc[i], ztc[i] = xyz_coord(Rtc[i], eps_tc[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xmp[i], ymp[i], zmp[i] = xyz_coord(Rmp[i], eps_mp[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xms[i], yms[i], zms[i] = xyz_coord(Rms[i], eps_ms[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xmc[i], ymc[i], zmc[i] = xyz_coord(Rmc[i], eps_mc[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xop[i], yop[i], zop[i] = xyz_coord(Rop[i], eps_op[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xos[i], yos[i], zos[i] = xyz_coord(Ros[i], eps_os[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])
            xoc[i], yoc[i], zoc[i] = xyz_coord(Roc[i], eps_oc[i], alpha[i],
                                               theta_sp[i], Ric[i], zic[i],
                                               delta[i])

        ## step 4 - calculate the blade section coordinates through blade
        # element stacking points (horizontal sections with constant x)
        ## [FR: etape 4 - calcul des coordonnees des sections a travers les
        # points d'empilement (sections horizontale avec x constant)]
        x = np.linspace(min(xsp), max(xsp), nx)

        # interpolate for a given number of levels the value of z for the
        # points
        # [FR: interpolation des valeurs de z sur un nombre donne de niveaux]
        for i in range(nx):
            zipx[i] = interpolation_extrapolation(xip, zipp, x[i])
            zisx[i] = interpolation_extrapolation(xis, zis, x[i])

            zopx[i] = interpolation_extrapolation(xop, zop, x[i])
            zosx[i] = interpolation_extrapolation(xos, zos, x[i])

            zicx[i] = interpolation_extrapolation(xic, zic, x[i])
            zocx[i] = interpolation_extrapolation(xoc, zoc, x[i])

        zmin = min((min(zipx),min(zisx)))
        zmax = max((max(zopx), max(zosx)))
        zx = np.linspace(zmin, zmax, nz)

        for j in range(nz):
            for i in range(nprofiles):
                Rz = Ric[i] + (zx[j] - zic[i]) / np.cos(alpha[i])
                if Rz > Rtc[i]:
                    kappa_c[i, j] = fix_arcsin(zeta_oc[i] - Rz * Coc[i])
                    epsilon_c[i, j] = eps_tc[i] + epsilon_variation(
                        kappa_c[i, j], ktc[i], zeta_oc[i], Rz, Rtc[i])
                else:
                    kappa_c[i, j] = fix_arcsin(zeta_ic[i] - Rz * Cic[i])
                    epsilon_c[i, j] = epsilon_variation(
                        kappa_c[i, j], kic[i], zeta_ic[i], Rz, Ric[i])
                if Rz > Rtp[i]:
                    kappa_p[i, j] = fix_arcsin(zeta_op[i] - Rz * Cop[i])
                    epsilon_p[i, j] = eps_tp[i] + epsilon_variation(
                        kappa_p[i, j], ktp[i], zeta_op[i], Rz, Rtp[i])
                else:
                    kappa_p[i, j] = fix_arcsin(zeta_ip[i] - Rz * Cip[i])
                    epsilon_p[i, j] = eps_ip[i] + epsilon_variation(
                        kappa_p[i, j], kip[i], zeta_ip[i], Rz, Rip[i])
                if Rz > Rts[i]:
                    kappa_s[i, j] = fix_arcsin(zeta_os[i] - Rz * Cos[i])
                    epsilon_s[i, j] = eps_ts[i] + epsilon_variation(
                        kappa_s[i, j], kts[i], zeta_os[i], Rz, Rts[i])
                else:
                    kappa_s[i, j] = fix_arcsin(zeta_is[i] - Rz * Cis[i])
                    epsilon_s[i, j] = eps_is[i] + epsilon_variation(
                        kappa_s[i, j], kis[i], zeta_is[i], Rz, Ris[i])

                xc[i, j], yc[i, j], temp = xyz_coord(
                    Rz, epsilon_c[i, j], alpha[i], theta_sp[i], Ric[i], zic[i],
                    delta[i])
                xs[i, j], ys[i, j], temp = xyz_coord(
                    Rz, epsilon_s[i, j], alpha[i], theta_sp[i], Ric[i], zic[i],
                    delta[i])
                xp[i, j], yp[i, j], temp = xyz_coord(
                    Rz, epsilon_p[i, j], alpha[i], theta_sp[i], Ric[i], zic[i],
                    delta[i])

        for k in range(nx):
            for j in range(nz):
                ypx[k, j] = interpolation_extrapolation(xp[:, j], yp[:, j],
                                                        x[k])
                ysx[k, j] = interpolation_extrapolation(xs[:, j], ys[:, j],
                                                        x[k])
                ycx[k, j] = interpolation_extrapolation(xc[:, j], yc[:, j],
                                                        x[k])
            yisx[k] = interpolation_extrapolation(zx, ysx[k,:], zisx[k])
            yosx[k] = interpolation_extrapolation(zx, ysx[k,:], zosx[k])
            yipx[k] = interpolation_extrapolation(zx, ypx[k,:], zipx[k])
            yopx[k] = interpolation_extrapolation(zx, ypx[k,:], zopx[k])
            yicx[k] = interpolation_extrapolation(zx, ycx[k, :], zicx[k])
            yocx[k] = interpolation_extrapolation(zx, ycx[k, :], zocx[k])

        ## step 5 - calculate the blade section centers of area
        # [FR: etape 5 - calcul du centre de gravite des sections]
        for k in range(nx):
            yspx[k], zspx[k] = center_gravity(yipx[k], zipx[k],
                                              yisx[k], zisx[k],
                                              ypx[k,:], ysx[k,:], zx,
                                              yopx[k], zopx[k],
                                              yosx[k], zosx[k])

        ## step 6 - calculate the new blade-element stacking point
        # and evaluate the differences between the new stacking points and the
        # previous stacking points
        # [FR: calcul des nouveaux points d'empilement et evaluation de la
        # difference entre les nouveaux points d'empilements et les points
        # d'empilement precedents]
        spdiff_norm = 0
        for i in range(nprofiles):
            # first approximation for new zsp and ysp
            # [FR: premiere approximation des nouveaux zsp et ysp]
            zspr[i] = interpolation_extrapolation(x, zspx, xsp[i])
            yspr[i] = interpolation_extrapolation(x, yspx, xsp[i])
            drsp = (zspr[i] - zsp[i]) * np.tan(alpha[i])
            # approximate xsp
            # [FR: approximation de xsp]
            xspr[i] = np.sqrt((rsp[i] + drsp) ** 2 - yspr[i] ** 2)
            # new zsp and ysp
            # [FR: nouveaux zsp et ysp]
            zspr[i] = interpolation_extrapolation(x, zspx, xspr[i])
            yspr[i] = interpolation_extrapolation(x, yspx, xspr[i])
            drsp = (zspr[i] - zsp[i]) * np.tan(alpha[i])
            # new xsp
            # [FR: nouveau xsp]
            xspr[i] = np.sqrt((rsp[i] + drsp) ** 2 - yspr[i] ** 2)
            # error computation, eq. (49)
            # [FR: calcul de l'erreur, eq. (49)]
            spdiff_norm = spdiff_norm + abs(zspr[i] - zsp[i]) +\
                          abs(yspr[i] - ysp[i])
        if last_step:
            # this was the last iteration after reaching the tolerance on
            # previous step, the iterations were successful
            # [FR: il s'agissait de la derniere iteration apres avoir atteint la
            # tolerance desiree a l'iteration precedente, les iterations sont
            # un succes]
            success = True
        elif spdiff_norm < tolerance:
            # tolerance reached, next step is the last step
            # [FR: tolerance atteinte, la prochaine iteration est la derniere]
            last_step = True
        else:
            # realignment of blade elements
            # [FR: realignement des profils]
            for i in range(nprofiles):
                theta_sp[i] = theta_sp[i] - delta[i] +\
                              np.arctan(yspr[i] / xspr[i])
                Rsp[i] = Ric[i] + (zspr[i] - zic[i]) / np.cos(alpha[i])
                rsp[i] = ric[i] + (zspr[i] - zic[i]) * np.tan(alpha[i])
                ## alternative formulation: identical result
                # rsp[i] = np.sqrt(xspr[i] ** 2 + yspr[i] ** 2)
            n += 1

    # list of profile coordinates
    # [FR: liste des coordonnees des profils]
    xlist = [[]] * nprofiles
    ylist = [[]] * nprofiles
    zlist = [[]] * nprofiles

    # number of points on the half-circles closing the profiles
    # [FR: nombre de points dans les demi-cercles fermants les profils]
    nc = 4
    for i in range(nprofiles):
        xlist[i] = []
        ylist[i] = []
        zlist[i] = []
        # suction side
        # [FR: extrados]
        index = np.where((zx >= zis[i]) & (zx <= zos[i]))
        xlist[i].extend(xs[i, index].flatten())
        ylist[i].extend(ys[i, index].flatten())
        zlist[i].extend(zx[index])
        # outlet semi-circle
        # [FR: demi-cercle du bord de fuite]
        for a in np.linspace(np.pi/2, -np.pi/2, nc):
            Rtemp, eps_temp = conical_coor(Roc[i], eps_oc[i], koc[i], to[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xlist[i].append(xtemp)
            ylist[i].append(ytemp)
            zlist[i].append(ztemp)
        # pressure side
        # [FR: intrados]
        index = np.where((zx >= zipp[i]) & (zx <= zop[i]))
        xlist[i].extend(xp[i, index].flatten()[::-1])
        ylist[i].extend(yp[i, index].flatten()[::-1])
        zlist[i].extend(zx[index][::-1])
        # inlet semi-circle
        # [FR: demi-cercle du bord d'attaque]
        for a in np.linspace(-np.pi/2,-3*np.pi/2,nc):
            Rtemp, eps_temp = conical_coor(Ric[i], eps_ic[i], kic[i], ti[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xlist[i].append(xtemp)
            ylist[i].append(ytemp)
            zlist[i].append(ztemp)

    # list of profile coordinates, suction surface
    # [FR: liste des coordonnees des profils, extrados]
    xslist = [[]] * nprofiles
    yslist = [[]] * nprofiles
    zslist = [[]] * nprofiles

    for i in range(nprofiles):
        xslist[i] = []
        yslist[i] = []
        zslist[i] = []
        # inlet quarter-circle
        # [FR: quart de cercle du bord d'attaque]
        for a in np.linspace(-np.pi, -3*np.pi/2, int(nc/2)):
            Rtemp, eps_temp = conical_coor(Ric[i], eps_ic[i], kic[i], ti[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xslist[i].append(xtemp)
            yslist[i].append(ytemp)
            zslist[i].append(ztemp)
        # suction side
        # [FR: extrados]
        index = np.where((zx >= zis[i]) & (zx <= zos[i]))
        xslist[i].extend(xs[i, index].flatten())
        yslist[i].extend(ys[i, index].flatten())
        zslist[i].extend(zx[index])
        # outlet quarter-circle
        # [FR: quart de cercle du bord de fuite]
        for a in np.linspace(np.pi/2, 0, int(nc/2)):
            Rtemp, eps_temp = conical_coor(Roc[i], eps_oc[i], koc[i], to[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xslist[i].append(xtemp)
            yslist[i].append(ytemp)
            zslist[i].append(ztemp)

    # list of profile coordinates, pressure surface
    # [FR: liste des coordonnees des profils, intrados]
    xplist = [[]] * nprofiles
    yplist = [[]] * nprofiles
    zplist = [[]] * nprofiles

    for i in range(nprofiles):
        xplist[i] = []
        yplist[i] = []
        zplist[i] = []
        # outlet quarter-circle
        # [FR: quart de cercle du bord de fuite]
        for a in np.linspace(0, -np.pi/2, int(nc/2)):
            Rtemp, eps_temp = conical_coor(Roc[i], eps_oc[i], koc[i], to[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xplist[i].append(xtemp)
            yplist[i].append(ytemp)
            zplist[i].append(ztemp)
        # pressure side
        # [FR: intrados]
        index = np.where((zx >= zipp[i]) & (zx <= zop[i]))
        xplist[i].extend(xp[i, index].flatten()[::-1])
        yplist[i].extend(yp[i, index].flatten()[::-1])
        zplist[i].extend(zx[index][::-1])
        # inlet quarter-circle
        # [FR: quart de cercle du bord d'attaque]
        for a in np.linspace(-np.pi/2,-np.pi,nc):
            Rtemp, eps_temp = conical_coor(Ric[i], eps_ic[i], kic[i], ti[i], a)
            xtemp, ytemp, ztemp = xyz_coord(Rtemp, eps_temp, alpha[i],
                                            theta_sp[i], Rsp[i], zsp[i],
                                            delta[i])
            xplist[i].append(xtemp)
            yplist[i].append(ytemp)
            zplist[i].append(ztemp)

    # list of leading edge and trailing edge coordinates
    # [FR: liste des coordonnees de bords d'attaque et bords de fuite]
    xle = [xslist[i][0] for i in range(nprofiles)]
    yle = [yslist[i][0] for i in range(nprofiles)]
    zle = [zslist[i][0] for i in range(nprofiles)]
    xte = [xslist[i][-1] for i in range(nprofiles)]
    yte = [yslist[i][-1] for i in range(nprofiles)]
    zte = [zslist[i][-1] for i in range(nprofiles)]
    for i in range(nprofiles):
        Rtemp, eps_temp = conical_coor(Ric[i], eps_ic[i], kic[i], ti[i], -np.pi)
        xle[i], yle[i], zle[i] = xyz_coord(Rtemp, eps_temp, alpha[i],
                                           theta_sp[i], Rsp[i], zsp[i],
                                           delta[i])
        Rtemp, eps_temp = conical_coor(Roc[i], eps_oc[i], koc[i], to[i], 0)
        xte[i], yte[i], zte[i] = xyz_coord(Rtemp, eps_temp, alpha[i],
                                           theta_sp[i], Rsp[i], zsp[i],
                                           delta[i])

    # plot the 3D geometry
    # [FR: representation graphique de la geometrie 3D]
    if display_fig:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for i in range(nprofiles):
            ax.plot(xlist[i], ylist[i], zlist[i], 'k:')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        ax.plot(xle, yle, zle, 'bo-')
        ax.plot(xte, yte, zte, 'ro-')
        ax.plot(xspr, yspr, zspr, 'ko-')

        # create cubic bounding box to simulate equal aspect ratio
        # [FR: creation d'une boite cubique pour simuler des echelles egales
        # en x, y et z]
        max_range = np.array([xc.max() - xc.min(),
                              yc.max() - yc.min(),
                              zc.max() - zc.min()]).max()
        Xb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][0].flatten() +\
             0.5 * (xc.max() + xc.min())
        Yb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][1].flatten() +\
             0.5 * (yc.max() + yc.min())
        Zb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][2].flatten() +\
             0.5 * (zc.max() + zc.min())
        for xb, yb, zb in zip(Xb, Yb, Zb):
            ax.plot([xb], [yb], [zb], 'w')

        plt.show()
    
    if save_profiles:
        if not os.path.exists(output_coor):
            os.makedirs(output_coor)
        for i, k in enumerate(percent_span):
            # suction side
            # [FR: extrados]
            tab = np.array([np.array(xslist[i]), np.array(yslist[i]),
                            np.array(zslist[i])])
            np.savetxt(os.path.join(output_coor,
                                    'profile_{}s.txt'.format(int(k))), tab.T)
            # pressure side
            # [FR: intrados]
            tab = np.array([np.array(xplist[i]), np.array(yplist[i]),
                            np.array(zplist[i])])
            np.savetxt(os.path.join(output_coor,
                                    'profile_{}p.txt'.format(int(k))), tab.T)
            # leading edges
            # [FR: bords d'attaque]
            tab = np.array([np.array(xle), np.array(yle), np.array(zle)])
            np.savetxt(os.path.join(output_coor,
                                    'profile_le.txt'), tab.T)
            # trailing edges
            # [FR: bords de fuite]
            tab = np.array([np.array(xte), np.array(yte), np.array(zte)])
            np.savetxt(os.path.join(output_coor,'profile_te.txt'), tab.T)
        percent_span = np.sort(percent_span)
        np.savetxt(os.path.join(output_coor, 'percent_span.txt'), percent_span,
                   fmt='%3i')


if __name__ == "__main__":
    #
    # retrieve parameters from ini file
    # [FR: recuperation des parametres a partir du fichier ini]
    #
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    script_path = os.path.dirname(os.path.abspath(filename))
    config = ConfigParser.ConfigParser()
    config.read(os.path.join(script_path, "..", "parameters.ini"))

    blade_name = config.get('profile_coordinates', 'blade_name')
    profile_file = os.path.join(script_path, "..", config.get('profile_coordinates',
                                                        'profile_folder'),
                                config.get('profile_coordinates',
                                           'profile_filename'))

    sweep_lean = config.getboolean('profile_coordinates', 'sweep_lean')
    if sweep_lean:
        sweep_file = os.path.join(script_path, "..", config.get('profile_coordinates',
                                                          'sweep_lean_folder'),
                                  blade_name + '_sweep.json')
        lean_file = os.path.join(script_path, "..", config.get('profile_coordinates',
                                                         'sweep_lean_folder'),
                                 blade_name + '_lean.json')
    else:
        sweep_file = ''
        lean_file = ''
    output_coor_directory = os.path.join(
        script_path, config.get('profile_coordinates', 'output_folder'),
        blade_name)
    display_fig = config.getboolean('profile_coordinates', 'display_fig')

    # launch the main program
    # [FR: lancement du programme principal]
    profile_coordinates(profile_file, output_coor_directory,
                        sweep_file, lean_file,display_fig=display_fig,
                        save_profiles=True)
